from __future__ import annotations

from copy import copy
from inspect import isclass
from typing import TypeVar

TOriginal = TypeVar("TOriginal")


def get_migration_import_class_placeholder(original: TOriginal, module: str) -> TOriginal:
    # get a placeholder type to clean up imported module paths in migrations

    original_type = original if isclass(original) else type(original)
    if not original_type.__module__.startswith(module):
        return original

    if not isclass(original):
        placeholder_type = get_migration_import_class_placeholder(type(original), module)
        copied = copy(original)
        copied.__class__ = placeholder_type
        return copied

    attr = f"_{original_type.__module__}.{original_type.__name__}"
    if hasattr(get_migration_import_class_placeholder, attr):
        return getattr(get_migration_import_class_placeholder, attr)

    copied_type = copy(original_type)
    copied_type.__module__ = module
    setattr(get_migration_import_class_placeholder, attr, copied_type)

    return copied_type



from __future__ import annotations, annotations

from typing import TYPE_CHECKING, Type, TypeVar, cast

from django.db.models import Field

from . import migration_helpers

if TYPE_CHECKING:
    from .json_model import JSONModel

TType = TypeVar("TType", bound=type)


class JSONModelState:
    """
    Used to represent the state of a JSONModel class in migrations to prevent importing the actual JSONModel subclass.
    """

    @classmethod
    def from_model(cls, model_class: Type[JSONModel]) -> JSONModelState:
        """
        Deconstructs a JSONModel class into JSONModelState instance that is used to represent the model in migrations.
        """

        bases = tuple(
            (
                cls._get_deconstructed_parent(parent)
                for parent in model_class.__bases__
            )
        )
        opts = model_class._meta
        fields = [(field.name, field) for field in opts.get_fields(include_parents=False)]
        attrs = {}
        if hasattr(model_class, "__classcell__"):
            attrs["__classcell__"] = model_class.__classcell__
        if opts.abstract:
            attrs["abstract"] = True

        state_kwargs = {
            "name": model_class.__name__,
            "bases": tuple(bases),
            "fields": fields,
        }
        if attrs:
            state_kwargs["attrs"] = attrs

        return JSONModelState(**state_kwargs)

    @classmethod
    def _get_deconstructed_parent(cls, parent: type) -> type | JSONModelState:
        from .json_model import JSONModel

        if parent == JSONModel:
            return migration_helpers.get_migration_import_class_placeholder(
                JSONModel,
                "django_json_model_field.db.models"
            )

        if issubclass(parent, JSONModel):
            return cls.from_model(parent)

        return parent

    def __init__(
        self,
        name: str,
        bases: tuple[type | JSONModelState, ...],
        fields: list[tuple[str, Field]],
        attrs: dict = None,
    ):
        self.name = name
        self.bases = bases
        self.fields = fields
        self.attrs = attrs or {}

    def deconstruct(self) -> tuple[str, tuple, dict]:
        attrs = {
            "name": self.name,
            "bases": self.bases,
            "fields": self.fields,
        }
        if self.attrs:
            attrs["attrs"] = self.attrs

        return "django_json_model_field.db.models.JSONModelState", (), attrs

    def reconstruct(self) -> Type[JSONModel]:
        """
        Reconstructs a JSONModel class from a deconstructed representation. Used when reconstructing JSONModelField
        instances from migration files during Django migration commands.

        The resulting classes are used for comparing the current state of the model as defined in the code against
        the cumulatively rendered state of the model in the migration files. It is not used at runtime.
        """

        bases = tuple(base.reconstruct() if isinstance(base, JSONModelState) else base for base in self.bases)
        attrs: dict = {name: field for name, field in self.fields}
        attrs.update(self.attrs)
        attrs.update(__module__="django_json_model_field.db.models.__reconstructed__")

        from .json_model import JSONModel

        return cast(Type[JSONModel], type(self.name, bases, attrs))


def reconstruct(state: JSONModelState | Type[JSONModel]) -> Type[JSONModel]:
    """
    Convenience function for reconstructing a JSONModel class from a JSONModelState instance or a JSONModel class.
    """

    return state.reconstruct() if isinstance(state, JSONModelState) else state



from __future__ import annotations

from abc import ABC, abstractmethod
from json import JSONDecoder, JSONEncoder
from typing import TYPE_CHECKING, Type

from django.db.models import Field

from . import migration_compat, migration_helpers

if TYPE_CHECKING:
    from .json_model import JSONModel


class MigrationCompatFieldMixin(Field, ABC):

    encoder: Type[JSONEncoder] | None
    decoder: Type[JSONDecoder] | None

    def __init__(
        self,
        *args,
        # internal flags used for deconstructing compatibility migrations
        # see comments in deconstruct_compat module for details
        _deconstruct_compat_migration: tuple[str, str, str] = None,
        _cloned: bool = False,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._deconstruct_compat_migration = self._init_deconstruct_compat_migration(
            _deconstruct_compat_migration,
            _cloned
        )

    @abstractmethod
    def _deconstruct_compat_is_reconstructed(self) -> bool:
        raise NotImplementedError()

    @classmethod
    def _deconstruct_compat_model_is_reconstructed(cls, model: Type[JSONModel]) -> bool:
        return model.__module__.endswith(".__reconstructed__")

    def deconstruct(self):
        return self._deconstruct_compat()

    def _deconstruct_compat(self, *, cloned: bool = False):
        """
        deconstruct() implementation that adds metadata to trigger a deconstruct compatibility migration if needed
        see comments in deconstruct_compat module for details
        """

        name, path, args, kwargs = super().deconstruct()

        if cloned:
            kwargs["_cloned"] = True

        kwargs["encoder"] = migration_helpers.get_migration_import_class_placeholder(
            self.encoder,
            "django_json_model_field.serialization"
        )
        kwargs["decoder"] = migration_helpers.get_migration_import_class_placeholder(
            self.decoder,
            "django_json_model_field.serialization"
        )

        if migration_compat.is_enabled() and not migration_compat.is_migrated_field(self):
            # cloned = coming from migration check to get the current state of the field
            if cloned and hasattr(self, "model"):
                kwargs["_deconstruct_compat_migration"] = migration_compat.get_migration_entry(self)
            elif self._deconstruct_compat_is_reconstructed() and self._deconstruct_compat_migration:
                kwargs["_deconstruct_compat_migration"] = self._deconstruct_compat_migration

        if path.startswith("django_json_model_field.db.models"):
            # provide the more concise path for the JSONModelField module in migrations - the import will use
            # django_json_model_field.db.models.JSONModelField
            # instead of
            # django_json_model_field.db.models.json_model_field.JSONModelField
            _, field_class_name = path.rsplit(".", 1)
            path = f"django_json_model_field.db.models.{field_class_name}"

        return name, path, args, kwargs

    def _init_deconstruct_compat_migration(
        self,
        deconstruct_compat_migration: tuple[tuple[str, str], str] | None,
        cloned: bool,
    ) -> tuple[tuple[str, str], str] | None:
        if not cloned and deconstruct_compat_migration:
            migration_compat.mark_migrated(deconstruct_compat_migration)

        return deconstruct_compat_migration

    def clone(self):
        name, path, args, kwargs = self._deconstruct_compat(cloned=True)
        return self.__class__(*args, **kwargs)



from __future__ import annotations

import os
from collections import defaultdict

from django.db.models import Field

# dict of (app_label, model_name) -> set of field names
_DECONSTRUCT_MIGRATED_FIELDS: dict[tuple[str, str], set[str]] = defaultdict(set)

"""
Utilities for tracking whether JSONModelFields have had a "compatibility migration" generated. The compatibility 
migration is needed for JSONModel fields that were created before the v1.0 release which added proper support for
deconstructing JSONModel fields for migrations.

Previously, JSONModel classes were imported directly into migrations, which causes problems when the JSONModel class
declaration changes or is moved. The compatibility migration creates a one-time AlterField operation that effectively
replaces the app state version of the field with one that uses correctly deconstructed JSONModel classes. This allows
subsequent changes to JSONModel classes to correctly generate migrations.

The compatibility migration is only generated once per field when running manage.py makemigrations with the environment
variable DJANGO_JSON_MODEL_FIELD_MIGRATION_COMPAT set to a truthy value.
"""


def is_enabled():
    return bool(os.environ.get("DJANGO_JSON_MODEL_FIELD_MIGRATION_COMPAT", False))


def mark_migrated(field: tuple[tuple[str, str], str]) -> None:
    model, field_name = field
    app_label, _ = model
    _DECONSTRUCT_MIGRATED_FIELDS[model].add(field_name)


def is_migrated_field(field: Field) -> bool:
    try:
        model, field = get_migration_entry(field)
        return model in _DECONSTRUCT_MIGRATED_FIELDS and field in _DECONSTRUCT_MIGRATED_FIELDS[model]
    except AttributeError:
        return False


def get_migration_entry(field: Field) -> tuple[[str, str], str]:
    entry = getattr(field, "_deconstruct_compat_migration", None)
    if entry:
        return entry
    return (field.model._meta.app_label or field.model.__module__, field.model._meta.model_name), field.name



from __future__ import annotations

from inspect import isclass
from typing import Type, overload

from django.apps.registry import Apps
from django.db.models import Field, Model

from django_json_model_field.db.models import (
    BaseJSONModelField,
    ConditionalJSONModelField,
    JSONModel,
    JSONModelDictField,
    JSONModelDictState,
)


@overload
def get_json_model_class(
    model: Type[Model] | Type[JSONModel],
    field_name: str,
    /,
    *,
    selector_value: str = None,
) -> Type[JSONModel]:
    ...


@overload
def get_json_model_class(
    apps: Apps,
    app_label: str,
    model_name: str,
    field_name: str,
    /,
    *,
    selector_value: str = None,
) -> Type[JSONModel]:
    ...


def get_json_model_class(
    apps_or_model: Apps | Type[Model] | Type[JSONModel],
    model_or_app_label_or_field_name: Type[Model] | Type[JSONModel] | str,
    field_or_model_name: str = None,
    field_name: str = None,
    /,
    *,
    selector_value: str = None,
) -> Type[JSONModel]:
    """
    Helper for loading JSONModel classes in migrations for use in RunPython operations.

    Usage:

        from django_json_model_field import migration_helpers


        def forwards_func(apps, schema_editor):
            # using a model class
            HostClass = apps.get_model("my_app", "HostClass")
            JSONModelClass = migration_helpers.get_json_model_class(apps, HostClass, "json_model_field")

            # or, using app_label and model_name
            JSONModelClass = get_json_model_class(apps, "my_app", "HostClass", "json_model_field")

            json_model_data = JSONModelClass(...)
            obj = HostClass.objects.create(json_model_field=json_model_data)

            # get_json_model_class can also be used with nested JSONModels
            NestedJSONModelClass = get_json_model_class(apps, JSONModelClass, "nested_json_model_field")
            nested_json_model_data = NestedJSONModelClass(...)

            obj.json_model_field.nested_json_model_field = nested_json_model_data
            obj.save()

    """

    model, field, selector_value = _get_field_from_args(
        apps_or_model=apps_or_model,
        model_or_app_label_or_field_name=model_or_app_label_or_field_name,
        field_or_model_name=field_or_model_name,
        field_name=field_name,
        selector_value=selector_value,
    )

    if isinstance(field, JSONModelDictField):
        raise TypeError(f"{field_name} is a JSONModelDictField, use get_json_model_dict_class instead")

    if isinstance(field, ConditionalJSONModelField):
        if selector_value is None:
            raise ValueError(f"{field_name} is a ConditionalJSONModelField and requires a selector_value")
        return field.get_json_model_class({field.selector_field_name: selector_value})

    if selector_value is not None:
        raise ValueError(f"{field_name} is not a ConditionalJSONModelField and does not support a selector_value")

    return field.get_json_model_class({})


@overload
def get_json_model_dict_class(
    model: Type[Model] | Type[JSONModel],
    field_name: str,
    /,
) -> tuple[Type[dict], Type[JSONModel]]:
    ...


@overload
def get_json_model_dict_class(
    apps: Apps,
    app_label: str,
    model_name: str,
    field_name: str,
    /,
) -> tuple[Type[dict], Type[JSONModel]]:
    ...


def get_json_model_dict_class(
    apps_or_model: Apps | Type[Model] | Type[JSONModel],
    model_or_app_label_or_field_name: Type[Model] | Type[JSONModel] | str,
    field_or_model_name: str = None,
    field_name: str = None,
    /,
) -> tuple[Type[dict[str, JSONModel]], Type[JSONModel]]:

    model, field, _ = _get_field_from_args(
        apps_or_model=apps_or_model,
        model_or_app_label_or_field_name=model_or_app_label_or_field_name,
        field_or_model_name=field_or_model_name,
        field_name=field_name,
    )

    if not isinstance(field, JSONModelDictField):
        raise TypeError(f"{field_name} is a not JSONModelDictField, use get_json_model_class instead")

    value_class = JSONModelDictState.get_value_class(field.json_dict_model)
    return field.json_dict_model, value_class


def _get_field_from_args(
    *,
    apps_or_model: Apps | Type[Model] | Type[JSONModel],
    model_or_app_label_or_field_name: Type[Model] | Type[JSONModel] | str,
    field_or_model_name: str | None,
    field_name: str | None,
    selector_value: str = None
) -> tuple[Type[Model] | Type[JSONModel], BaseJSONModelField | JSONModelDictField, str | None]:
    if isclass(apps_or_model):
        if not issubclass(apps_or_model, (Model, JSONModel)):
            raise TypeError("model must be a Django Model or JSONModel class")
        model = apps_or_model
        field_name = model_or_app_label_or_field_name
    elif isclass(model_or_app_label_or_field_name):
        if not issubclass(model_or_app_label_or_field_name, (Model, JSONModel)):
            raise TypeError("model must be a Django Model or JSONModel class")
        model = model_or_app_label_or_field_name
        field_name = field_or_model_name
    else:
        if not isinstance(apps_or_model, Apps):
            raise TypeError("apps must be an Apps instance")
        if not isinstance(model_or_app_label_or_field_name, str):
            raise TypeError("app_label must be a string")
        if not isinstance(field_or_model_name, str):
            raise TypeError("model_name must be a string")
        apps = apps_or_model
        app_label = model_or_app_label_or_field_name
        model_name = field_or_model_name
        model = apps.get_model(app_label, model_name)

    if not isinstance(field_name, str):
        raise TypeError("field_name must be a string")

    if selector_value is not None and not isinstance(selector_value, str):
        raise TypeError("selector_value must be a string")

    field = _get_field(model, field_name)

    if not isinstance(field, (BaseJSONModelField, JSONModelDictField)):
        raise TypeError(f"{field_name} is not a JSONModel field")

    return model, field, selector_value


def _get_field(model: Type[Model] | Type[JSONModel], field_name: str) -> Field:
    return next(field for field in model._meta.get_fields() if field.name == field_name)



from .base_json_model_field import BaseJSONModelField
from .conditional_json_model_field import ConditionalJSONModelField
from .json_model import JSONModel
from .json_model_dict_field import JSONModelDictField
from .json_model_dict_state import JSONModelDictState
from .json_model_field import JSONModelField
from .json_model_state import JSONModelState
from .migration_compat_field_mixin import MigrationCompatFieldMixin



from __future__ import annotations, annotations

from importlib import import_module
from types import new_class
from typing import Any, Dict, TYPE_CHECKING, Type, TypeVar, cast, overload

from .json_model_state import JSONModelState

if TYPE_CHECKING:
    from .json_model import JSONModel

TType = TypeVar("TType", bound=type)


class JSONModelDictState:
    """
    Used to represent the state of a JSONModel class in migrations to prevent importing the actual JSONModel subclass.
    """

    @classmethod
    def from_model(cls, model_class: Type[dict[Any, JSONModel]]) -> JSONModelDictState:
        """
        Deconstructs a JSONModelDict class into JSONModelDictState instance that is used to represent the model in
        migrations.
        """

        is_typing_dict = str(model_class).startswith("typing.Dict")
        dict_class = dict if is_typing_dict else model_class
        dict_class_name = "dict" if is_typing_dict else getattr(dict_class, "__qualname__", dict_class.__name__)

        value_class = _get_value_class(model_class)

        from .json_model import JSONModel

        if not issubclass(value_class, (JSONModel, dict)):
            raise TypeError(f"Value class {value_class} must be a subclass of JSONModel")

        state_kwargs = {
            "dict_class": (dict_class.__module__, dict_class_name),
            "value_class": (
                JSONModelDictState.from_model(value_class)
                if issubclass(value_class, dict)
                else JSONModelState.from_model(value_class)
            ),
        }

        return JSONModelDictState(**state_kwargs)

    @classmethod
    def get_value_class(cls, model_class: Type[dict[Any, JSONModel]]) -> Type[JSONModel]:
        return _get_value_class(model_class)

    def __init__(
        self,
        dict_class: tuple[str, str],
        value_class: JSONModelState,
    ):
        self.dict_class = dict_class
        self.value_class = value_class

    def deconstruct(self) -> tuple[str, tuple, dict]:
        attrs = {
            "dict_class": self.dict_class,
            "value_class": self.value_class,
        }

        return "django_json_model_field.db.models.JSONModelDictState", (), attrs

    def reconstruct(self) -> Type[dict]:
        """
        Reconstructs a JSONModel class from a deconstructed representation. Used when reconstructing JSONModelField
        instances from migration files during Django migration commands.

        The resulting classes are used for comparing the current state of the model as defined in the code against
        the cumulatively rendered state of the model in the migration files. It is not used at runtime.
        """

        dict_class_module, dict_class_name = self.dict_class
        dict_class = (
            Dict if (dict_class_module, dict_class_name) == ("builtins", "dict")
            else getattr(import_module(dict_class_module), dict_class_name)
        )
        value_class = self.value_class.reconstruct()

        result = new_class(dict_class_name, (dict_class,))
        result.__module__ = dict_class_module
        result.__args__ = (str, value_class)

        return cast(Type[dict], result)


def reconstruct(state: JSONModelDictState | Type[dict]) -> Type[dict]:
    """
    Convenience function for reconstructing a JSONModel class from a JSONModelState instance or a JSONModel class.
    """

    return state.reconstruct() if isinstance(state, JSONModelDictState) else state


@overload
def _get_value_class(model_class: type) -> Type[JSONModel]:
    ...


@overload
def _get_value_class(model_class: type, *, _depth: int = 0) -> Type[JSONModel] | None:
    ...


def _get_value_class(model_class: type, *, _depth: int = 0) -> Type[JSONModel] | None:

    value_class = _get_value_arg(getattr(model_class, "__args__", ()))
    if value_class is not None:
        from .json_model import JSONModel

        if issubclass(value_class, JSONModel):
            return value_class

        # JSONModelDict may be nested
        if issubclass(value_class, dict):
            return _get_value_class(value_class, _depth=_depth + 1)

    value_class = _find_value_class_in_bases(getattr(model_class, "__orig_bases__", ()), _depth=_depth)
    if value_class is not None:
        return value_class

    value_class = _find_value_class_in_bases(getattr(model_class, "__bases__", ()), _depth=_depth)
    if value_class is not None:
        return value_class

    if _depth > 0:
        return None

    raise TypeError(
        f"Model class {model_class} must be a generic dict type with a value type that is a JSONModel subclass "
        f"or another JSONModelDict type"
    )


def _get_value_arg(args: tuple[type, ...]) -> type | None:
    return len(args) == 2 and args[1] or None


def _find_value_class_in_bases(bases: tuple[type, ...], *, _depth: int) -> Type[JSONModel] | None:
    for base in bases:
        base_result = _get_value_class(base, _depth=_depth + 1)
        if base_result is not None:
            return base_result
    return None
